import hashlib
import itertools
import re
from functools import lru_cache

from django.template.loader import render_to_string
from django.urls import reverse

from wagtail import hooks

icon_comment_pattern = re.compile(r"<!--.*?-->", re.DOTALL)


@lru_cache(maxsize=None)
def get_icons():
    icon_hooks = hooks.get_hooks("register_icons")
    all_icons = sorted(itertools.chain.from_iterable(hook([]) for hook in icon_hooks))
    combined_icon_markup = ""
    for icon in all_icons:
        symbol = (
            render_to_string(icon)
            .replace('xmlns="http://www.w3.org/2000/svg"', "")
            .replace("svg", "symbol")
        )
        symbol = icon_comment_pattern.sub("", symbol)
        combined_icon_markup += symbol

    return render_to_string(
        "wagtailadmin/shared/icons.html", {"icons": combined_icon_markup}
    )


@lru_cache(maxsize=None)
def get_icon_sprite_hash():
    return hashlib.sha1(get_icons().encode()).hexdigest()[:8]


def get_icon_sprite_url():
    return reverse("wagtailadmin_sprite") + f"?h={get_icon_sprite_hash()}"



# Sendfile "streaming" backend
# This is based on sendfiles builtin "simple" backend but uses a StreamingHttpResponse

import os
import stat
from email.utils import mktime_tz, parsedate_tz

from django.http import FileResponse, HttpResponseNotModified
from django.utils.http import http_date


def sendfile(request, filename, **kwargs):
    # Respect the If-Modified-Since header.
    statobj = os.stat(filename)

    if not was_modified_since(
        request.headers.get("if-modified-since"),
        statobj[stat.ST_MTIME],
    ):
        return HttpResponseNotModified()

    response = FileResponse(open(filename, "rb"))

    response["Last-Modified"] = http_date(statobj[stat.ST_MTIME])
    return response


def was_modified_since(header=None, mtime=0):
    """
    Was something modified since the user last downloaded it?

    header
      This is the value of the If-Modified-Since header.  If this is None,
      I'll just return True.

    mtime
      This is the modification time of the item we're talking about.
    """
    try:
        if header is None:
            raise ValueError
        header_date = parsedate_tz(header)
        if header_date is None:
            raise ValueError
        header_mtime = mktime_tz(header_date)
        if mtime > header_mtime:
            raise ValueError
    except (ValueError, OverflowError):
        return True
    return False



from wagtail.documents import get_document_model_string
from wagtail.permission_policies.collections import CollectionOwnershipPermissionPolicy

permission_policy = CollectionOwnershipPermissionPolicy(
    get_document_model_string(),
    auth_model="wagtaildocs.Document",
    owner_field_name="uploaded_by_user",
)



from django.urls import path

from wagtail.admin.views import account

urlpatterns = [
    path("", account.PasswordResetView.as_view(), name="wagtailadmin_password_reset"),
    path(
        "done/",
        account.PasswordResetDoneView.as_view(),
        name="wagtailadmin_password_reset_done",
    ),
    path(
        "confirm/<uidb64>/<token>/",
        account.PasswordResetConfirmView.as_view(),
        name="wagtailadmin_password_reset_confirm",
    ),
    path(
        "complete/",
        account.PasswordResetCompleteView.as_view(),
        name="wagtailadmin_password_reset_complete",
    ),
]



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailusers", "0002_add_verbose_name_on_userprofile"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="userprofile",
            options={"verbose_name": "User Profile"},
        ),
    ]



from django.urls import include, path, reverse_lazy
from django.utils.translation import gettext_lazy as _

import wagtail.admin.rich_text.editors.draftail.features as draftail_features
from wagtail import hooks
from wagtail.embeds import urls
from wagtail.embeds.rich_text import MediaEmbedHandler
from wagtail.embeds.rich_text.contentstate import ContentstateMediaConversionRule
from wagtail.embeds.rich_text.editor_html import EditorHTMLEmbedConversionRule


@hooks.register("register_admin_urls")
def register_admin_urls():
    return [
        path("embeds/", include(urls, namespace="wagtailembeds")),
    ]


@hooks.register("register_rich_text_features")
def register_embed_feature(features):
    # define a handler for converting <embed embedtype="media"> tags into frontend HTML
    features.register_embed_type(MediaEmbedHandler)

    # define how to convert between editorhtml's representation of embeds and
    # the database representation
    features.register_converter_rule(
        "editorhtml", "embed", EditorHTMLEmbedConversionRule
    )

    # define a draftail plugin to use when the 'embed' feature is active
    features.register_editor_plugin(
        "draftail",
        "embed",
        draftail_features.EntityFeature(
            {
                "type": "EMBED",
                "icon": "media",
                "description": _("Embed"),
                "chooserUrls": {
                    "embedsChooser": reverse_lazy("wagtailembeds:chooser"),
                },
            },
            js=["wagtailembeds/js/embed-chooser-modal.js"],
        ),
    )

    # define how to convert between contentstate's representation of embeds and
    # the database representation-
    features.register_converter_rule(
        "contentstate", "embed", ContentstateMediaConversionRule
    )

    # add 'embed' to the set of on-by-default rich text features
    features.default_features.append("embed")



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import F
from django.forms import CheckboxSelectMultiple, RadioSelect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.functional import cached_property, classproperty
from django.utils.translation import gettext_lazy as _
from django_filters.filters import (
    ChoiceFilter,
    DateFromToRangeFilter,
    ModelMultipleChoiceFilter,
)

from wagtail import hooks
from wagtail.admin.filters import (
    DateRangePickerWidget,
    MultipleContentTypeFilter,
    MultipleUserFilter,
    WagtailFilterSet,
)
from wagtail.admin.ui.components import MediaContainer
from wagtail.admin.ui.side_panels import (
    PageStatusSidePanel,
)
from wagtail.admin.ui.tables import DateColumn
from wagtail.admin.ui.tables.pages import (
    BulkActionsColumn,
    NavigateToChildrenColumn,
    OrderingColumn,
    PageStatusColumn,
    PageTable,
    PageTitleColumn,
    PageTypeColumn,
    ParentPageColumn,
)
from wagtail.admin.views import generic
from wagtail.models import Page, PageLogEntry, Site, get_page_content_types
from wagtail.permissions import page_permission_policy


class SiteFilter(ModelMultipleChoiceFilter):
    def get_filter_predicate(self, v):
        return {"path__startswith": v.root_page.path}


class HasChildPagesFilter(ChoiceFilter):
    def filter(self, qs, value):
        if value == "true":
            return qs.filter(numchild__gt=0)
        elif value == "false":
            return qs.filter(numchild=0)
        else:  # None / empty string
            return qs


class EditedByFilter(MultipleUserFilter):
    def filter(self, qs, value):
        if value:
            qs = qs.filter(
                pk__in=PageLogEntry.objects.filter(
                    action="wagtail.edit", user__in=value
                )
                .order_by()
                .values_list("page_id", flat=True)
                .distinct()
            )
        return qs


class PageFilterSet(WagtailFilterSet):
    latest_revision_created_at = DateFromToRangeFilter(
        label=_("Date updated"),
        widget=DateRangePickerWidget,
    )
    owner = MultipleUserFilter(
        label=_("Owner"),
        queryset=(
            lambda request: get_user_model().objects.filter(
                pk__in=Page.objects.order_by()
                .values_list("owner_id", flat=True)
                .distinct()
            )
        ),
        widget=CheckboxSelectMultiple,
    )
    edited_by = EditedByFilter(
        label=_("Edited by"),
        queryset=(
            lambda request: get_user_model().objects.filter(
                pk__in=PageLogEntry.objects.filter(action="wagtail.edit")
                .order_by()
                .values_list("user_id", flat=True)
                .distinct()
            )
        ),
        widget=CheckboxSelectMultiple,
    )
    site = SiteFilter(
        label=_("Site"),
        queryset=Site.objects.all(),
        widget=CheckboxSelectMultiple,
    )
    has_child_pages = HasChildPagesFilter(
        label=_("Has child pages"),
        empty_label=_("Any"),
        choices=[
            ("true", _("Yes")),
            ("false", _("No")),
        ],
        widget=RadioSelect,
    )

    class Meta:
        model = Page
        fields = []  # only needed for filters being generated automatically


class GenericPageFilterSet(PageFilterSet):
    content_type = MultipleContentTypeFilter(
        label=_("Page type"),
        queryset=lambda request: get_page_content_types(include_base_page_type=False),
        widget=CheckboxSelectMultiple,
    )


class PageListingMixin:
    template_name = "wagtailadmin/pages/listing.html"
    context_object_name = "pages"
    table_class = PageTable
    filterset_class = GenericPageFilterSet
    default_ordering = "-latest_revision_created_at"
    model = Page
    is_searchable = True

    columns = [
        BulkActionsColumn("bulk_actions"),
        PageTitleColumn(
            "title",
            label=_("Title"),
            sort_key="title",
            classname="title",
        ),
        ParentPageColumn("parent", label=_("Parent")),
        DateColumn(
            "latest_revision_created_at",
            label=_("Updated"),
            sort_key="latest_revision_created_at",
            width="12%",
        ),
        PageTypeColumn(
            "type",
            label=_("Type"),
            accessor="page_type_display_name",
            sort_key="content_type",
            width="12%",
        ),
        PageStatusColumn(
            "status",
            label=_("Status"),
            sort_key="live",
            width="12%",
        ),
    ]

    @cached_property
    def i18n_enabled(self):
        return getattr(settings, "WAGTAIL_I18N_ENABLED", False)

    @cached_property
    def show_locale_labels(self):
        return self.i18n_enabled

    def get_valid_orderings(self):
        valid_orderings = super().get_valid_orderings()

        if self.is_searching:
            # ordering by content type not currently available when searching, due to
            # https://github.com/wagtail/wagtail/issues/6616
            try:
                valid_orderings.remove("content_type")
                valid_orderings.remove("-content_type")
            except ValueError:
                pass

        return valid_orderings

    def get_ordering(self):
        if self.is_searching and not self.is_explicitly_ordered:
            # default to ordering by relevance
            default_ordering = None
        else:
            default_ordering = self.default_ordering

        ordering = self.request.GET.get("ordering", default_ordering)
        if ordering not in self.get_valid_orderings():
            ordering = default_ordering

        return ordering

    def annotate_queryset(self, pages):
        pages = pages.prefetch_related("content_type", "sites_rooted_here")

        # We want specific page instances, but do not need streamfield values here
        pages = pages.defer_streamfields().specific()

        # Annotate queryset with various states to be used later for performance optimisations
        if getattr(settings, "WAGTAIL_WORKFLOW_ENABLED", True):
            pages = pages.prefetch_workflow_states()
        if self.i18n_enabled:
            pages = pages.prefetch_related("locale").annotate_has_untranslated_locale()

        pages = pages.annotate_site_root_state().annotate_approved_schedule()

        return pages

    def order_queryset(self, queryset):
        if self.is_searching and not self.is_explicitly_ordered:
            # search backend will order by relevance in this case, so don't bother to
            # apply an ordering on the queryset
            return queryset

        if self.ordering == "ord":
            # preserve the native ordering from get_children()
            pass
        elif self.ordering == "latest_revision_created_at" and not self.is_searching:
            # order by oldest revision first.
            # Special case NULL entries - these should go at the top of the list.
            # Skip this special case when searching (and fall through to plain field ordering
            # instead) as search backends do not support F objects in order_by
            queryset = queryset.order_by(
                F("latest_revision_created_at").asc(nulls_first=True)
            )
        elif self.ordering == "-latest_revision_created_at" and not self.is_searching:
            # order by oldest revision first.
            # Special case NULL entries - these should go at the end of the list.
            # Skip this special case when searching (and fall through to plain field ordering
            # instead) as search backends do not support F objects in order_by
            queryset = queryset.order_by(
                F("latest_revision_created_at").desc(nulls_last=True)
            )
        else:
            queryset = super().order_queryset(queryset)

        return queryset

    def search_queryset(self, queryset):
        if self.is_searching:
            queryset = queryset.autocomplete(
                self.search_query, order_by_relevance=(not self.is_explicitly_ordered)
            )

        return queryset

    def get_table_kwargs(self):
        kwargs = super().get_table_kwargs()
        kwargs["actions_next_url"] = self.index_url
        kwargs["show_locale_labels"] = self.show_locale_labels
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        if any(isinstance(column, ParentPageColumn) for column in self.columns):
            Page.objects.annotate_parent_page(context["object_list"])

        return context


class IndexView(PageListingMixin, generic.IndexView):
    permission_policy = page_permission_policy
    any_permission_required = {
        "add",
        "change",
        "publish",
        "bulk_delete",
        "lock",
        "unlock",
    }
    template_name = "wagtailadmin/pages/index.html"
    results_template_name = "wagtailadmin/pages/index_results.html"
    paginate_by = 50
    table_classname = "listing full-width"
    filterset_class = PageFilterSet

    @classproperty
    def columns(cls):
        return [col for col in PageListingMixin.columns if col.name != "type"]

    def get_base_queryset(self):
        pages = self.model.objects.filter(depth__gt=1).filter(
            pk__in=page_permission_policy.explorable_instances(
                self.request.user
            ).values_list("pk", flat=True)
        )
        pages = self.annotate_queryset(pages)
        return pages


class ExplorableIndexView(IndexView):
    """
    A version of the page listing where the user is presented with a view of a specified parent page;
    normally this will show the children of that page, but it may show results from the whole tree while
    searching or filtering.
    """

    template_name = "wagtailadmin/pages/explorable_index.html"
    results_template_name = "wagtailadmin/pages/explorable_index_results.html"
    index_url_name = "wagtailadmin_explore"
    index_results_url_name = "wagtailadmin_explore_results"
    page_title = _("Exploring")
    filterset_class = GenericPageFilterSet

    @classproperty
    def columns(cls):
        columns = [col for col in PageListingMixin.columns if col.name != "parent"]
        columns.append(NavigateToChildrenColumn("navigate", width="10%"))
        return columns

    def get(self, request, parent_page_id=None):
        if parent_page_id:
            self.parent_page = get_object_or_404(
                Page.objects.all().prefetch_workflow_states(), id=parent_page_id
            )
        else:
            self.parent_page = Page.get_first_root_node()

        # This will always succeed because of the check performed by PermissionCheckedMixin.
        root_page = self.permission_policy.explorable_root_instance(request.user)

        # If this page isn't a descendant of the user's explorable root page,
        # then redirect to that explorable root page instead.
        if not (
            self.parent_page.pk == root_page.pk
            or self.parent_page.is_descendant_of(root_page)
        ):
            return redirect(self.index_url_name, root_page.pk)

        self.parent_page = self.parent_page.specific
        self.scheduled_page = self.parent_page.get_scheduled_revision_as_object()

        if self.i18n_enabled and not self.parent_page.is_root():
            self.locale = self.parent_page.locale
            self.translations = self.get_translations()
        else:
            self.locale = None
            self.translations = []

        return super().get(request)

    @cached_property
    def is_searching_whole_tree(self):
        return bool(self.request.GET.get("search_all")) and (
            self.is_searching or self.is_filtering
        )

    @cached_property
    def show_locale_labels(self):
        return self.i18n_enabled and self.parent_page.is_root()

    def get_base_queryset(self):
        if self.is_searching or self.is_filtering:
            if self.is_searching_whole_tree:
                pages = Page.objects.all()
            else:
                pages = self.parent_page.get_descendants()
        else:
            pages = self.parent_page.get_children()

        pages = pages.filter(
            pk__in=page_permission_policy.explorable_instances(
                self.request.user
            ).values_list("pk", flat=True)
        )
        pages = self.annotate_queryset(pages)
        return pages

    def search_queryset(self, queryset):
        # allow hooks to modify queryset. This should happen as close as possible to the
        # final queryset, but (for backward compatibility) needs to be passed an actual queryset
        # rather than a search result object
        for hook in hooks.get_hooks("construct_explorer_page_queryset"):
            queryset = hook(self.parent_page, queryset, self.request)
        return super().search_queryset(queryset)

    def get_index_url(self):
        return reverse(self.index_url_name, args=[self.parent_page.id])

    def get_index_results_url(self):
        return reverse(self.index_results_url_name, args=[self.parent_page.id])

    def get_history_url(self):
        permissions = self.parent_page.permissions_for_user(self.request.user)
        if permissions.can_view_revisions():
            return reverse("wagtailadmin_pages:history", args=[self.parent_page.id])

    def get_table_kwargs(self):
        kwargs = super().get_table_kwargs()
        kwargs["use_row_ordering_attributes"] = self.show_ordering_column
        kwargs["parent_page"] = self.parent_page

        if self.show_ordering_column:
            kwargs["caption"] = _(
                "Focus on the drag button and press up or down arrows to move the item, then press enter to submit the change."
            )
            kwargs["attrs"] = {
                "data-controller": "w-orderable",
                "data-w-orderable-active-class": "w-orderable--active",
                "data-w-orderable-chosen-class": "w-orderable__item--active",
                "data-w-orderable-container-value": "tbody",
                "data-w-orderable-message-value": _(
                    "'%(page_title)s' has been moved successfully."
                )
                % {"page_title": "__LABEL__"},
                "data-w-orderable-url-value": reverse(
                    "wagtailadmin_pages:set_page_position", args=[999999]
                ),
            }
        return kwargs

    def get_valid_orderings(self):
        valid_orderings = super().get_valid_orderings()

        if not self.is_searching:
            # ordering by page order is only available when not searching
            valid_orderings.append("ord")

        return valid_orderings

    def get_ordering(self):
        if self.is_searching and not self.is_explicitly_ordered:
            # default to ordering by relevance
            default_ordering = None
        else:
            default_ordering = self.parent_page.get_admin_default_ordering()

        ordering = self.request.GET.get("ordering", default_ordering)
        if ordering not in self.get_valid_orderings():
            ordering = default_ordering

        return ordering

    def get_paginate_by(self, queryset):
        if self.ordering == "ord":
            # Don't paginate if sorting by page order - all pages must be shown to
            # allow drag-and-drop reordering
            return None
        else:
            return self.paginate_by

    def get_page_subtitle(self):
        return self.parent_page.get_admin_display_title()

    def get_context_data(self, **kwargs):
        self.show_ordering_column = self.ordering == "ord"
        if self.show_ordering_column:
            self.columns = self.columns.copy()
            self.columns[0] = OrderingColumn("ordering", width="80px", sort_key="ord")
        context = super().get_context_data(**kwargs)

        if self.is_searching:
            Page.objects.annotate_parent_page(context["object_list"])
            for page in context["object_list"]:
                # add annotation if parent page is found and is not the currently viewed parent
                # to be used by PageTitleColumn instead of a dedicated ParentPageColumn
                if page._parent_page and page._parent_page != self.parent_page:
                    page.annotated_parent_page = page._parent_page

        context.update(
            {
                "parent_page": self.parent_page,
                "history_url": self.get_history_url(),
                "is_searching_whole_tree": self.is_searching_whole_tree,
            }
        )

        if not self.results_only:
            side_panels = self.get_side_panels()
            context["side_panels"] = side_panels
            context["media"] += side_panels.media

        return context

    def get_side_panels(self):
        # Don't show side panels on the root page
        if self.parent_page.is_root():
            return MediaContainer()

        side_panels = [
            PageStatusSidePanel(
                self.parent_page.get_latest_revision_as_object(),
                self.request,
                show_schedule_publishing_toggle=False,
                live_object=self.parent_page,
                scheduled_object=self.scheduled_page,
                locale=self.locale,
                translations=self.translations,
            ),
        ]
        return MediaContainer(side_panels)

    def get_translations(self):
        return [
            {
                "locale": translation.locale,
                "url": reverse(self.index_url_name, args=[translation.id]),
            }
            for translation in self.parent_page.get_translations()
            .only("id", "locale")
            .select_related("locale")
        ]



from wagtail.images import get_image_model
from wagtail.images.formats import get_image_format
from wagtail.rich_text import EmbedHandler

# Front-end conversion


class ImageEmbedHandler(EmbedHandler):
    identifier = "image"

    @staticmethod
    def get_model():
        return get_image_model()

    @classmethod
    def expand_db_attributes(cls, attrs: dict) -> str:
        return cls.expand_db_attributes_many([attrs])[0]

    @classmethod
    def expand_db_attributes_many(cls, attrs_list: list[dict]) -> list[str]:
        """
        Given a dict of attributes from the <embed> tag, return the real HTML
        representation for use on the front-end.
        """
        images = cls.get_many(attrs_list)

        tags = []
        for attrs, image in zip(attrs_list, images):
            if image:
                image_format = get_image_format(attrs["format"])
                tag = image_format.image_to_html(image, attrs.get("alt", ""))
            else:
                tag = '<img alt="">'

            tags.append(tag)

        return tags

    @classmethod
    def extract_references(cls, attrs):
        # Yields tuples of (content_type_id, object_id, model_path, content_path)
        yield cls.get_model(), attrs["id"], "", ""



import itertools
import os
import re
from collections import defaultdict

from django import forms
from django.core.paginator import Paginator
from django.template.loader import render_to_string
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView

from wagtail import hooks
from wagtail.admin import messages
from wagtail.admin.forms.search import SearchForm
from wagtail.admin.rich_text import get_rich_text_editor_widget
from wagtail.admin.staticfiles import versioned_static
from wagtail.admin.views.generic import WagtailAdminTemplateMixin
from wagtail.admin.widgets import (
    AdminAutoHeightTextInput,
    AdminDateInput,
    AdminDateTimeInput,
    AdminPageChooser,
    AdminTimeInput,
    SwitchInput,
)
from wagtail.compat import URLField
from wagtail.documents.widgets import AdminDocumentChooser
from wagtail.images.widgets import AdminImageChooser
from wagtail.models import Page
from wagtail.snippets.widgets import AdminSnippetChooser


class FakeAdminSnippetChooser(AdminSnippetChooser):
    """
    AdminSnippetChooser can't be used on non-snippet models (because it fails when constructing the
    URL to the chooser modal), and we can't guarantee that any given Wagtail installation using
    this style guide will have any snippet models registered. We therefore override the
    get_chooser_modal_url method so that we can use it with Page as a stand-in for a real snippet.
    """

    def get_chooser_modal_url(self):
        return "/"


class ExampleForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["page_chooser"].widget = AdminPageChooser()
        self.fields["image_chooser"].widget = AdminImageChooser()
        self.fields["document_chooser"].widget = AdminDocumentChooser()
        self.fields["snippet_chooser"].widget = FakeAdminSnippetChooser(Page)
        self.fields["date"].widget = AdminDateInput()
        self.fields["time"].widget = AdminTimeInput()
        self.fields["datetime"].widget = AdminDateTimeInput()
        self.fields["auto_height_text"].widget = AdminAutoHeightTextInput()
        self.fields["default_rich_text"].widget = get_rich_text_editor_widget("default")
        self.fields["switch"].widget = SwitchInput()
        self.fields["disabled_switch"].widget = SwitchInput(attrs={"disabled": True})

    CHOICES = (
        ("choice1", "choice 1"),
        ("choice2", "choice 2 that is longer but pretty normal"),
        (
            "choice3",
            """
            choice 3 that has a very long label that it cannot possibly fit within the
            width of the parent container but we're going to test it anyway to see
            what happens and how it wraps and whether it breaks the layout or not
            """,
        ),
    )

    text = forms.CharField(required=True, help_text="help text")
    auto_height_text = forms.CharField(required=True)
    default_rich_text = forms.CharField(required=True)
    url = URLField(required=True)
    email = forms.EmailField(max_length=254)
    date = forms.DateField()
    time = forms.TimeField()
    datetime = forms.DateTimeField()
    select = forms.ChoiceField(choices=CHOICES[:2])
    long_select = forms.ChoiceField(choices=CHOICES)
    radio_select = forms.ChoiceField(choices=CHOICES, widget=forms.RadioSelect)
    multiple_select = forms.MultipleChoiceField(choices=CHOICES)
    multiple_checkbox = forms.MultipleChoiceField(
        choices=CHOICES, widget=forms.CheckboxSelectMultiple
    )
    boolean = forms.BooleanField(required=False)
    switch = forms.BooleanField(required=False)
    disabled_switch = forms.BooleanField(required=False)
    page_chooser = forms.BooleanField(required=True)
    image_chooser = forms.BooleanField(required=True)
    document_chooser = forms.BooleanField(required=True)
    snippet_chooser = forms.BooleanField(required=True)

    @property
    def media(self):
        return super().media + forms.Media(
            css={
                "all": [versioned_static("wagtailstyleguide/css/animate-progress.css")]
            }
        )


# Allow single and double quotes for the ID.
# For simplicity and readability, we don't enforce the opening
# and closing quotes to match.
icon_id_pattern = re.compile(r"""id=["']icon-([a-z0-9-]+)["']""")
icon_comment_pattern = re.compile(r"<!--!(.*?)-->")


class IndexView(WagtailAdminTemplateMixin, TemplateView):
    template_name = "wagtailstyleguide/base.html"
    page_title = _("Styleguide")
    header_icon = "image"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        messages.success(
            self.request,
            _("Success message"),
            buttons=[
                messages.button("", _("View live")),
                messages.button("", _("Edit")),
            ],
        )
        messages.warning(
            self.request,
            _("Warning message"),
            buttons=[
                messages.button("", _("View live")),
                messages.button("", _("Edit")),
            ],
        )
        messages.error(
            self.request,
            _("Error message"),
            buttons=[
                messages.button("", _("View live")),
                messages.button("", _("Edit")),
            ],
        )

        context["all_icons"] = self.get_icons()
        context["example_form"] = ExampleForm()
        context["example_page"] = Paginator(list(range(100)), 10).page(2)
        context["search_form"] = SearchForm()

        return context

    def get_icons(self):
        icon_hooks = hooks.get_hooks("register_icons")
        registered_icons = itertools.chain.from_iterable(
            hook([]) for hook in icon_hooks
        )
        all_icons = defaultdict(list)
        for icon_path in registered_icons:
            folder, filename = os.path.split(icon_path)
            icon = render_to_string(icon_path)
            id_match = icon_id_pattern.search(icon)
            name = id_match.group(1) if id_match else None
            source_match = icon_comment_pattern.search(icon)

            all_icons[folder].append(
                {
                    "folder": folder,
                    "file_path": icon_path,
                    "name": name,
                    "source": source_match.group(1) if source_match else None,
                }
            )
        return all_icons.items()



import django_filters
from django import forms
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.core.paginator import Paginator
from django.db import transaction
from django.db.models import Count, Prefetch
from django.db.models.functions import Lower
from django.http import Http404, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, redirect, render
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext
from django.views.decorators.http import require_POST
from django.views.generic import TemplateView

from wagtail.admin import messages
from wagtail.admin.auth import PermissionPolicyChecker
from wagtail.admin.filters import (
    MultipleContentTypeFilter,
    WagtailFilterSet,
)
from wagtail.admin.forms.workflows import (
    TaskChooserSearchForm,
    WorkflowContentTypeForm,
    WorkflowPagesFormSet,
    get_task_form_class,
    get_workflow_edit_handler,
)
from wagtail.admin.modal_workflow import render_modal_workflow
from wagtail.admin.ui.tables import BaseColumn, Column, TitleColumn
from wagtail.admin.views.generic import CreateView, DeleteView, EditView, IndexView
from wagtail.admin.views.generic.base import BaseListingView
from wagtail.admin.views.generic.permissions import PermissionCheckedMixin
from wagtail.admin.views.pages.listing import PageListingMixin
from wagtail.coreutils import resolve_model_string
from wagtail.models import (
    Page,
    Task,
    TaskState,
    Workflow,
    WorkflowState,
    WorkflowTask,
)
from wagtail.permissions import (
    page_permission_policy,
    task_permission_policy,
    workflow_permission_policy,
)
from wagtail.snippets.models import get_workflow_enabled_models
from wagtail.workflows import get_task_types

task_permission_checker = PermissionPolicyChecker(task_permission_policy)


class WorkflowTitleColumn(TitleColumn):
    cell_template_name = "wagtailadmin/workflows/includes/workflow_title_cell.html"


class WorkflowUsedByColumn(TitleColumn):
    cell_template_name = "wagtailadmin/workflows/includes/workflow_used_by_cell.html"

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["workflow_enabled_models"] = get_workflow_enabled_models()
        return context


class WorkflowTasksColumn(BaseColumn):
    cell_template_name = "wagtailadmin/workflows/includes/workflow_tasks_cell.html"
    num_tasks = 5

    def get_cell_context_data(self, instance, parent_context):
        context = super().get_cell_context_data(instance, parent_context)
        context["tasks"] = instance.workflow_tasks.all()[: self.num_tasks]
        context["extra_count"] = instance.workflow_tasks.count() - self.num_tasks
        return context


class BaseWorkflowFilterSet(WagtailFilterSet):
    show_disabled = django_filters.ChoiceFilter(
        label=_("Show disabled"),
        method="filter_show_disabled",
        choices=(("true", _("Yes")), ("false", _("No"))),
        widget=forms.RadioSelect,
        empty_label=None,
        initial="false",
    )

    def __init__(self, data=None, queryset=None, *, request=None, prefix=None):
        if data is not None:
            if data.get("show_disabled") is None:
                filter = self.base_filters["show_disabled"]
                data = data.copy()
                data["show_disabled"] = filter.extra["initial"]
        super().__init__(data, queryset, request=request, prefix=prefix)

    def filter_show_disabled(self, queryset, name, value):
        if value == "true":
            return queryset
        return queryset.filter(active=True)


class WorkflowFilterSet(BaseWorkflowFilterSet):
    class Meta:
        model = Workflow
        fields = []


class Index(IndexView):
    permission_policy = workflow_permission_policy
    model = Workflow
    context_object_name = "workflows"
    template_name = "wagtailadmin/workflows/index.html"
    results_template_name = "wagtailadmin/workflows/index_results.html"
    add_url_name = "wagtailadmin_workflows:add"
    edit_url_name = "wagtailadmin_workflows:edit"
    index_url_name = "wagtailadmin_workflows:index"
    index_results_url_name = "wagtailadmin_workflows:index_results"
    page_title = _("Workflows")
    add_item_label = _("Add a workflow")
    header_icon = "tasks"
    columns = [
        WorkflowTitleColumn(
            "name",
            label=_("Name"),
            url_name="wagtailadmin_workflows:edit",
            width="25%",
            sort_key="name",
        ),
        WorkflowUsedByColumn(
            "usage",
            label=_("Used by"),
            url_name="wagtailadmin_workflows:usage",
            width="15%",
        ),
        WorkflowTasksColumn("tasks", label=_("Tasks")),
    ]
    default_ordering = "name"
    search_fields = ["name"]
    filterset_class = WorkflowFilterSet
    paginate_by = 20

    def show_disabled(self):
        return self.filters.form.cleaned_data.get("show_disabled") == "true"

    def get_base_queryset(self):
        queryset = super().get_base_queryset()
        queryset = queryset.annotate(content_types=Count("workflow_content_types"))
        return queryset.prefetch_related(
            "workflow_pages",
            "workflow_pages__page",
            "workflow_tasks",
            "workflow_tasks__task",
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["showing_disabled"] = self.show_disabled()
        return context


class Create(CreateView):
    permission_policy = workflow_permission_policy
    model = Workflow
    page_title = _("New workflow")
    template_name = "wagtailadmin/workflows/create.html"
    success_message = _("Workflow '%(object)s' created.")
    add_url_name = "wagtailadmin_workflows:add"
    edit_url_name = "wagtailadmin_workflows:edit"
    index_url_name = "wagtailadmin_workflows:index"
    header_icon = "tasks"
    edit_handler = None

    def get_edit_handler(self):
        if not self.edit_handler:
            self.edit_handler = get_workflow_edit_handler()
        return self.edit_handler

    def get_form_class(self):
        return self.get_edit_handler().get_form_class()

    def get_pages_formset(self):
        if self.request.method == "POST":
            return WorkflowPagesFormSet(
                self.request.POST, instance=self.object, prefix="pages"
            )
        else:
            return WorkflowPagesFormSet(instance=self.object, prefix="pages")

    def get_content_type_form(self):
        if self.request.method == "POST":
            return WorkflowContentTypeForm(self.request.POST, workflow=self.object)
        else:
            return WorkflowContentTypeForm(workflow=self.object)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        form = context["form"]
        bound_panel = self.edit_handler.get_bound_panel(
            form=form, instance=form.instance, request=self.request
        )
        pages_formset = self.get_pages_formset()

        context["edit_handler"] = bound_panel
        context["pages_formset"] = pages_formset
        context["has_workflow_enabled_models"] = bool(get_workflow_enabled_models())
        context["content_type_form"] = self.get_content_type_form()
        context["media"] = form.media + bound_panel.media + pages_formset.media
        return context

    def form_valid(self, form):
        self.form = form

        with transaction.atomic():
            self.object = self.save_instance()

            pages_formset = self.get_pages_formset()
            content_type_form = self.get_content_type_form()
            if pages_formset.is_valid() and content_type_form.is_valid():
                pages_formset.save()
                content_type_form.save()

                success_message = self.get_success_message(self.object)
                if success_message is not None:
                    messages.success(
                        self.request,
                        success_message,
                        buttons=[
                            messages.button(
                                reverse(self.edit_url_name, args=(self.object.id,)),
                                _("Edit"),
                            )
                        ],
                    )
                return redirect(self.get_success_url())

            else:
                transaction.set_rollback(True)

        return self.form_invalid(form)


class Edit(EditView):
    permission_policy = workflow_permission_policy
    model = Workflow
    page_title = _("Editing workflow")
    template_name = "wagtailadmin/workflows/edit.html"
    success_message = _("Workflow '%(object)s' updated.")
    add_url_name = "wagtailadmin_workflows:add"
    edit_url_name = "wagtailadmin_workflows:edit"
    delete_url_name = "wagtailadmin_workflows:disable"
    delete_item_label = _("Disable")
    index_url_name = "wagtailadmin_workflows:index"
    enable_item_label = _("Enable")
    enable_url_name = "wagtailadmin_workflows:enable"
    header_icon = "tasks"
    header_more_buttons = []
    edit_handler = None
    MAX_PAGES = 5

    def get_edit_handler(self):
        if not self.edit_handler:
            self.edit_handler = get_workflow_edit_handler()
        return self.edit_handler

    def get_form_class(self):
        return self.get_edit_handler().get_form_class()

    def get_pages_formset(self):
        if self.request.method == "POST":
            return WorkflowPagesFormSet(
                self.request.POST, instance=self.get_object(), prefix="pages"
            )
        else:
            return WorkflowPagesFormSet(instance=self.get_object(), prefix="pages")

    def get_content_type_form(self):
        if self.request.method == "POST":
            return WorkflowContentTypeForm(self.request.POST, workflow=self.object)
        else:
            return WorkflowContentTypeForm(workflow=self.object)

    def get_paginated_pages(self):
        # Get the (paginated) list of Pages to which this Workflow is assigned.
        pages = Page.objects.filter(workflowpage__workflow=self.get_object())
        pages.paginator = Paginator(pages, self.MAX_PAGES)
        page_number = int(self.request.GET.get("p", 1))
        paginated_pages = pages.paginator.page(page_number)
        return paginated_pages

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        form = context["form"]
        bound_panel = self.edit_handler.get_bound_panel(
            form=form, instance=form.instance, request=self.request
        )
        pages_formset = self.get_pages_formset()
        context["edit_handler"] = bound_panel
        context["pages"] = self.get_paginated_pages()
        context["has_workflow_enabled_models"] = bool(get_workflow_enabled_models())
        context["content_type_form"] = self.get_content_type_form()
        context["can_disable"] = (
            self.permission_policy is None
            or self.permission_policy.user_has_permission(self.request.user, "delete")
        ) and self.object.active
        context["can_enable"] = (
            self.permission_policy is None
            or self.permission_policy.user_has_permission(self.request.user, "add")
        ) and not self.object.active
        context["media"] = bound_panel.media + form.media

        # Only add the pages_formset if the workflow is active
        if self.object.active:
            pages_formset = self.get_pages_formset()
            context["pages_formset"] = pages_formset
            context["media"] += pages_formset.media

        return context

    @property
    def get_enable_url(self):
        return reverse(self.enable_url_name, args=(self.object.pk,))

    @transaction.atomic()
    def form_valid(self, form):
        self.form = form

        with transaction.atomic():
            self.object = self.save_instance()
            successful = True

            # Save pages formset and content type form
            # Note: These are hidden when the workflow is inactive
            if self.object.active:
                pages_formset = self.get_pages_formset()
                content_type_form = self.get_content_type_form()
                if pages_formset.is_valid() and content_type_form.is_valid():
                    pages_formset.save()
                    content_type_form.save()
                else:
                    transaction.set_rollback(True)
                    successful = False

            if successful:
                success_message = self.get_success_message()
                if success_message is not None:
                    messages.success(
                        self.request,
                        success_message,
                        buttons=[
                            messages.button(
                                reverse(self.edit_url_name, args=(self.object.id,)),
                                _("Edit"),
                            )
                        ],
                    )
                return redirect(self.get_success_url())

        return self.form_invalid(form)


class Disable(DeleteView):
    permission_policy = workflow_permission_policy
    model = Workflow
    page_title = _("Disable workflow")
    template_name = "wagtailadmin/workflows/confirm_disable.html"
    success_message = _("Workflow '%(object)s' disabled.")
    add_url_name = "wagtailadmin_workflows:add"
    edit_url_name = "wagtailadmin_workflows:edit"
    delete_url_name = "wagtailadmin_workflows:disable"
    index_url_name = "wagtailadmin_workflows:index"
    header_icon = "tasks"

    @property
    def get_edit_url(self):
        return reverse(self.edit_url_name, args=(self.kwargs["pk"],))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        states_in_progress = WorkflowState.objects.filter(
            workflow=self.object, status=WorkflowState.STATUS_IN_PROGRESS
        ).count()
        if states_in_progress:
            context["warning_message"] = ngettext(
                "This workflow is in progress on %(states_in_progress)d page/snippet. Disabling this workflow will cancel moderation on this page/snippet.",
                "This workflow is in progress on %(states_in_progress)d pages/snippets. Disabling this workflow will cancel moderation on these pages/snippets.",
                states_in_progress,
            ) % {
                "states_in_progress": states_in_progress,
            }
        return context

    def delete_action(self):
        self.object.deactivate(user=self.request.user)


class WorkflowUsageView(PageListingMixin, PermissionCheckedMixin, BaseListingView):
    permission_policy = workflow_permission_policy
    any_permission_required = {"add", "change", "delete", "view"}
    pk_url_kwarg = "pk"
    index_url_name = "wagtailadmin_workflows:usage"
    index_results_url_name = "wagtailadmin_workflows:usage_results"
    paginate_by = 20
    header_icon = "tasks"
    page_title = _("Usage")

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()

        # We set workflow_permission_policy as the main permission policy for this view
        # for consistency with the other workflow views. However, since we are listing
        # page objects in this view, we want to ensure the user has a page permission.
        if not page_permission_policy.user_has_any_permission(
            request.user,
            {"add", "change", "publish", "bulk_delete", "lock", "unlock"},
        ):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)

    def get_page_subtitle(self):
        return self.object.name

    def get_breadcrumbs_items(self):
        title = self.get_page_title()
        subtitle = self.get_page_subtitle()
        return self.breadcrumbs_items + [
            {
                "url": reverse("wagtailadmin_workflows:index"),
                "label": capfirst(Workflow._meta.verbose_name_plural),
            },
            {
                "url": reverse("wagtailadmin_workflows:edit", args=(self.object.pk,)),
                "label": subtitle,
            },
            {"url": "", "label": title, "sublabel": subtitle},
        ]

    def get_index_url(self):
        return reverse(self.index_url_name, args=(self.object.pk,))

    def get_index_results_url(self):
        return reverse(self.index_results_url_name, args=(self.object.pk,))

    def get_object(self):
        return get_object_or_404(Workflow, id=self.kwargs.get(self.pk_url_kwarg))

    def get_base_queryset(self):
        editable_pages = page_permission_policy.instances_user_has_permission_for(
            self.request.user, "change"
        ).filter(depth__gt=1)
        pages = self.object.all_pages() & editable_pages
        pages = self.annotate_queryset(pages)
        return pages


@require_POST
def enable_workflow(request, pk):
    # Reactivate an inactive workflow
    workflow = get_object_or_404(Workflow, id=pk)

    # Check permissions
    if not workflow_permission_policy.user_has_permission(request.user, "add"):
        raise PermissionDenied

    # Set workflow to active if inactive
    if not workflow.active:
        workflow.active = True
        workflow.save()
        messages.success(
            request,
            _("Workflow '%(workflow_name)s' enabled.")
            % {"workflow_name": workflow.name},
        )

    # Redirect
    redirect_to = request.POST.get("next", None)
    if redirect_to and url_has_allowed_host_and_scheme(
        url=redirect_to, allowed_hosts={request.get_host()}
    ):
        return redirect(redirect_to)
    else:
        return redirect("wagtailadmin_workflows:edit", workflow.id)


@require_POST
def remove_workflow(request, page_pk, workflow_pk=None):
    # Remove a workflow from a page (specifically a single workflow if workflow_pk is set)
    # Get the page
    page = get_object_or_404(Page, id=page_pk)

    # Check permissions
    if not workflow_permission_policy.user_has_permission(request.user, "change"):
        raise PermissionDenied

    if hasattr(page, "workflowpage"):
        # If workflow_pk is set, this will only remove the workflow if it its pk matches - this prevents accidental
        # removal of the wrong workflow via a workflow edit page if the page listing is out of date
        if not workflow_pk or workflow_pk == page.workflowpage.workflow.pk:
            page.workflowpage.delete()
            messages.success(
                request,
                _("Workflow removed from Page '%(page_title)s'.")
                % {"page_title": page.get_admin_display_title()},
            )

    # Redirect
    redirect_to = request.POST.get("next", None)
    if redirect_to and url_has_allowed_host_and_scheme(
        url=redirect_to, allowed_hosts={request.get_host()}
    ):
        return redirect(redirect_to)
    else:
        return redirect("wagtailadmin_explore", page.id)


class TaskTitleColumn(TitleColumn):
    cell_template_name = "wagtailadmin/workflows/includes/task_title_cell.html"


class TaskUsageColumn(Column):
    cell_template_name = "wagtailadmin/workflows/includes/task_usage_cell.html"


class TaskFilterSet(BaseWorkflowFilterSet):
    def __init__(self, data=None, queryset=None, *, request=None, prefix=None):
        super().__init__(data, queryset, request=request, prefix=prefix)
        task_types = get_task_types()
        ct_ids = [
            ct.id for ct in ContentType.objects.get_for_models(*task_types).values()
        ]
        if len(task_types) > 1:
            self.filters["content_type"] = MultipleContentTypeFilter(
                label=_("Type"),
                widget=forms.CheckboxSelectMultiple,
                queryset=lambda request: ContentType.objects.filter(pk__in=ct_ids),
                field_name="content_type",
            )

    class Meta:
        model = Task
        fields = []


class TaskIndex(IndexView):
    permission_policy = task_permission_policy
    model = Task
    context_object_name = "tasks"
    template_name = "wagtailadmin/workflows/task_index.html"
    results_template_name = "wagtailadmin/workflows/task_index_results.html"
    add_url_name = "wagtailadmin_workflows:select_task_type"
    edit_url_name = "wagtailadmin_workflows:edit_task"
    index_url_name = "wagtailadmin_workflows:task_index"
    index_results_url_name = "wagtailadmin_workflows:task_index_results"
    page_title = _("Workflow tasks")
    add_item_label = _("New workflow task")
    header_icon = "thumbtack"
    columns = [
        TaskTitleColumn(
            "name",
            label=_("Name"),
            url_name="wagtailadmin_workflows:edit_task",
            sort_key="name",
        ),
        Column("type", label=_("Type"), accessor="get_verbose_name", width="25%"),
        TaskUsageColumn(
            "usage", label=_("Used on"), accessor="_active_workflows", width="25%"
        ),
    ]
    default_ordering = "name"
    search_fields = ["name"]
    filterset_class = TaskFilterSet
    paginate_by = 50

    def show_disabled(self):
        return self.filters.form.cleaned_data.get("show_disabled") == "true"

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .specific()
            .prefetch_related(
                Prefetch(
                    "workflow_tasks",
                    queryset=WorkflowTask.objects.filter(
                        workflow__active=True
                    ).select_related("workflow"),
                    to_attr="_active_workflows",
                )
            )
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["showing_disabled"] = self.show_disabled()
        return context


def select_task_type(request):
    if not task_permission_policy.user_has_permission(request.user, "add"):
        raise PermissionDenied

    task_types = [
        (
            model.get_verbose_name(),
            model._meta.app_label,
            model._meta.model_name,
            model.get_description(),
        )
        for model in get_task_types()
    ]
    # sort by lower-cased version of verbose name
    task_types.sort(key=lambda task_type: task_type[0].lower())

    if len(task_types) == 1:
        # Only one task type is available - redirect straight to the create form rather than
        # making the user choose
        verbose_name, app_label, model_name, description = task_types[0]
        return redirect("wagtailadmin_workflows:add_task", app_label, model_name)

    return render(
        request,
        "wagtailadmin/workflows/select_task_type.html",
        {
            "task_types": task_types,
            "icon": "thumbtack",
            "title": _("Workflows"),
        },
    )


class CreateTask(CreateView):
    permission_policy = task_permission_policy
    page_title = _("New workflow task")
    template_name = "wagtailadmin/workflows/create_task.html"
    success_message = _("Task '%(object)s' created.")
    add_url_name = "wagtailadmin_workflows:add_task"
    edit_url_name = "wagtailadmin_workflows:edit_task"
    index_url_name = "wagtailadmin_workflows:task_index"
    header_icon = "thumbtack"

    @cached_property
    def model(self):
        try:
            content_type = ContentType.objects.get_by_natural_key(
                self.kwargs["app_label"], self.kwargs["model_name"]
            )
        except (ContentType.DoesNotExist, AttributeError):
            raise Http404

        # Get class
        model = content_type.model_class()

        # Make sure the class is a descendant of Task
        if not issubclass(model, Task) or model is Task:
            raise Http404

        return model

    def get_form_class(self):
        return get_task_form_class(self.model)

    def get_add_url(self):
        return reverse(
            self.add_url_name,
            kwargs={
                "app_label": self.kwargs.get("app_label"),
                "model_name": self.kwargs.get("model_name"),
            },
        )

    def get_breadcrumbs_items(self):
        # Use the base Task class instead of the specific class for the index view
        items = [
            {
                "url": reverse(self.index_url_name),
                "label": capfirst(Task._meta.verbose_name_plural),
            },
            {
                "url": "",
                "label": _("New: %(model_name)s")
                % {"model_name": capfirst(self.model._meta.verbose_name)},
            },
        ]
        return self.breadcrumbs_items + items


class EditTask(EditView):
    permission_policy = task_permission_policy
    template_name = "wagtailadmin/workflows/edit_task.html"
    success_message = _("Task '%(object)s' updated.")
    add_url_name = "wagtailadmin_workflows:select_task_type"
    edit_url_name = "wagtailadmin_workflows:edit_task"
    delete_url_name = "wagtailadmin_workflows:disable_task"
    index_url_name = "wagtailadmin_workflows:task_index"
    delete_item_label = _("Disable")
    enable_item_label = _("Enable")
    enable_url_name = "wagtailadmin_workflows:enable_task"
    header_icon = "thumbtack"
    header_more_buttons = []

    @cached_property
    def model(self):
        return type(self.get_object())

    @cached_property
    def page_title(self):
        return _("Editing %(task_type)s") % {
            "task_type": self.get_object().content_type.name
        }

    def get_queryset(self):
        if self.queryset is None:
            return Task.objects.all()

    def get_object(self, queryset=None):
        return super().get_object().specific

    def get_form_class(self):
        return get_task_form_class(self.model, for_edit=True)

    def get_breadcrumbs_items(self):
        # Use the base Task class instead of the specific class
        items = [
            {
                "url": reverse(self.index_url_name),
                "label": capfirst(Task._meta.verbose_name_plural),
            },
            {"url": "", "label": str(self.object)},
        ]
        return self.breadcrumbs_items + items

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["can_disable"] = (
            self.permission_policy is None
            or self.permission_policy.user_has_permission(self.request.user, "delete")
        ) and self.object.active
        context["can_enable"] = (
            self.permission_policy is None
            or self.permission_policy.user_has_permission(self.request.user, "add")
        ) and not self.object.active

        # TODO: add warning msg when there are pages/snippets currently on this task in a workflow, add interaction like resetting task state when saved
        return context

    @property
    def get_enable_url(self):
        return reverse(self.enable_url_name, args=(self.object.pk,))


class DisableTask(DeleteView):
    permission_policy = task_permission_policy
    model = Task
    page_title = _("Disable task")
    template_name = "wagtailadmin/workflows/confirm_disable_task.html"
    success_message = _("Task '%(object)s' disabled.")
    add_url_name = "wagtailadmin_workflows:add_task"
    edit_url_name = "wagtailadmin_workflows:edit_task"
    delete_url_name = "wagtailadmin_workflows:disable_task"
    index_url_name = "wagtailadmin_workflows:task_index"
    header_icon = "thumbtack"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        states_in_progress = TaskState.objects.filter(
            status=TaskState.STATUS_IN_PROGRESS, task=self.get_object().pk
        ).count()
        if states_in_progress:
            context["warning_message"] = ngettext(
                "This task is in progress on %(states_in_progress)d page/snippet. Disabling this task will cause it to be skipped in the moderation workflow and not be listed for selection when editing a workflow.",
                "This task is in progress on %(states_in_progress)d pages/snippets. Disabling this task will cause it to be skipped in the moderation workflow and not be listed for selection when editing a workflow.",
                states_in_progress,
            ) % {
                "states_in_progress": states_in_progress,
            }
        return context

    @property
    def get_edit_url(self):
        return reverse(self.edit_url_name, args=(self.kwargs["pk"],))

    def delete_action(self):
        self.object.deactivate(user=self.request.user)


@require_POST
def enable_task(request, pk):
    # Reactivate an inactive task
    task = get_object_or_404(Task, id=pk)

    # Check permissions
    if not task_permission_policy.user_has_permission(request.user, "add"):
        raise PermissionDenied

    # Set workflow to active if inactive
    if not task.active:
        task.active = True
        task.save()
        messages.success(
            request, _("Task '%(task_name)s' enabled.") % {"task_name": task.name}
        )

    # Redirect
    redirect_to = request.POST.get("next", None)
    if redirect_to and url_has_allowed_host_and_scheme(
        url=redirect_to, allowed_hosts={request.get_host()}
    ):
        return redirect(redirect_to)
    else:
        return redirect("wagtailadmin_workflows:edit_task", task.id)


def get_task_chosen_response(request, task):
    """
    helper function: given a task, return the response indicating that it has been chosen
    """
    result_data = {
        "id": task.id,
        "name": task.name,
        "edit_url": reverse("wagtailadmin_workflows:edit_task", args=[task.id]),
    }
    return render_modal_workflow(
        request,
        None,
        None,
        None,
        json_data={"step": "task_chosen", "result": result_data},
    )


class BaseTaskChooserView(TemplateView):
    def dispatch(self, request):
        self.task_models = get_task_types()
        self.can_create = (
            task_permission_policy.user_has_permission(request.user, "add")
            and len(self.task_models) != 0
        )
        return super().dispatch(request)

    def get_create_model(self):
        """
        To be called after dispatch(); returns the model to use for a new task if one is known
        (either from being the only available task mode, or from being specified in the URL as create_model)
        """
        if self.can_create:
            if len(self.task_models) == 1:
                return self.task_models[0]

            elif "create_model" in self.request.GET:
                create_model = resolve_model_string(self.request.GET["create_model"])

                if create_model not in self.task_models:
                    raise Http404

                return create_model

    def get_create_form_class(self):
        """
        To be called after dispatch(); returns the form class for creating a new task
        """
        self.create_model = self.get_create_model()
        if self.create_model:
            return get_task_form_class(self.create_model)
        else:
            return None

    def get_create_form(self):
        """
        To be called after dispatch(); returns a blank create form, or None if not available
        """
        create_form_class = self.get_create_form_class()
        if create_form_class:
            return create_form_class(prefix="create-task")

    def get_task_type_options(self):
        """
        To be called after dispatch(); returns the task types list for the "select task type" view
        """
        task_types = [
            (
                model.get_verbose_name(),
                model._meta.app_label,
                model._meta.model_name,
                model.get_description(),
            )
            for model in self.task_models
        ]
        # sort by lower-cased version of verbose name
        task_types.sort(key=lambda task_type: task_type[0].lower())

        return task_types

    def get_task_type_filter_choices(self):
        """
        To be called after dispatch(); returns the list of task type choices for filter on "existing task" tab
        """
        task_type_choices = [
            (model, model.get_verbose_name()) for model in self.task_models
        ]
        task_type_choices.sort(key=lambda task_type: task_type[1].lower())
        return task_type_choices

    def get_form_js_context(self):
        return {}

    def get_task_listing_context_data(self):
        search_form = TaskChooserSearchForm(
            self.request.GET, task_type_choices=self.get_task_type_filter_choices()
        )
        tasks = all_tasks = search_form.task_model.objects.filter(active=True).order_by(
            Lower("name")
        )
        q = ""

        if search_form.is_searching():
            # Note: I decided not to use wagtailsearch here. This is because
            # wagtailsearch creates a new index for each model you make
            # searchable and this might affect someone's quota. I doubt there
            # would ever be enough tasks to require using anything more than
            # an icontains anyway.
            q = search_form.cleaned_data["q"]
            tasks = tasks.filter(name__icontains=q)

        # Pagination
        paginator = Paginator(tasks, per_page=10)
        tasks = paginator.get_page(self.request.GET.get("p"))

        return {
            "search_form": search_form,
            "tasks": tasks,
            "all_tasks": all_tasks,
            "query_string": q,
            "can_create": self.can_create,
        }

    def get_create_tab_context_data(self):
        return {
            "create_form": self.create_form,
            "add_url": reverse("wagtailadmin_workflows:task_chooser_create")
            + "?"
            + self.request.GET.urlencode()
            if self.create_model
            else None,
            "task_types": self.get_task_type_options(),
        }


class TaskChooserView(BaseTaskChooserView):
    def get(self, request):
        self.create_form = self.get_create_form()
        return super().get(request)

    def get_context_data(self, **kwargs):
        context = {
            "can_create": self.can_create,
        }
        context.update(self.get_task_listing_context_data())
        context.update(self.get_create_tab_context_data())
        return context

    def render_to_response(self, context):
        js_context = self.get_form_js_context()
        js_context["step"] = "chooser"

        return render_modal_workflow(
            self.request,
            "wagtailadmin/workflows/task_chooser/chooser.html",
            None,
            context,
            json_data=js_context,
        )


class TaskChooserCreateView(BaseTaskChooserView):
    def get(self, request):
        self.create_form = self.get_create_form()
        return super().get(request)

    def post(self, request):
        create_form_class = self.get_create_form_class()
        if not create_form_class:
            return HttpResponseBadRequest()

        self.create_form = create_form_class(
            request.POST, request.FILES, prefix="create-task"
        )

        if self.create_form.is_valid():
            task = self.create_form.save()
            return get_task_chosen_response(request, task)
        else:
            context = self.get_context_data()
            return self.render_to_response(context)

    def get_context_data(self, **kwargs):
        return self.get_create_tab_context_data()

    def render_to_response(self, context):
        tab_html = render_to_string(
            "wagtailadmin/workflows/task_chooser/includes/create_tab.html",
            context,
            self.request,
        )

        js_context = self.get_form_js_context()
        js_context["step"] = "reshow_create_tab"
        js_context["htmlFragment"] = tab_html

        return render_modal_workflow(
            self.request, None, None, None, json_data=js_context
        )


class TaskChooserResultsView(BaseTaskChooserView):
    template_name = "wagtailadmin/workflows/task_chooser/includes/results.html"

    def get_context_data(self, **kwargs):
        return self.get_task_listing_context_data()


def task_chosen(request, task_id):
    task = get_object_or_404(Task, id=task_id)
    return get_task_chosen_response(request, task)



# Generated by Django 3.0.10 on 2020-12-20 01:55

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("wagtaildocs", "0011_add_choose_permissions"),
    ]

    operations = [
        migrations.CreateModel(
            name="UploadedDocument",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "file",
                    models.FileField(max_length=200, upload_to="uploaded_documents"),
                ),
                (
                    "uploaded_by_user",
                    models.ForeignKey(
                        blank=True,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="uploaded by user",
                    ),
                ),
            ],
        ),
    ]



from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from wagtail.images.models import Image
from wagtail.images.tests.utils import get_test_image_file
from wagtail.permission_policies import (
    AuthenticationOnlyPermissionPolicy,
    BlanketPermissionPolicy,
    ModelPermissionPolicy,
    OwnershipPermissionPolicy,
)
from wagtail.test.utils import WagtailTestUtils


class PermissionPolicyTestUtils:
    def assertResultSetEqual(self, actual, expected):
        self.assertEqual(set(actual), set(expected))

    def assertUserPermissionMatrix(self, test_cases, actions=()):
        """
        Given a list of (user, can_add, can_change, can_delete, can_frobnicate) tuples
        (where 'frobnicate' is an unrecognised action not defined on the model),
        confirm that all tuples correctly represent permissions for that user as
        returned by user_has_permission
        """
        if not actions:
            actions = ["add", "change", "delete", "frobnicate"]
        for test_case in test_cases:
            user = test_case[0]
            expected_results = zip(actions, test_case[1:])

            for action, expected_result in expected_results:
                if expected_result:
                    self.assertTrue(
                        self.policy.user_has_permission(user, action),
                        msg="User {} should be able to {}, but can't".format(
                            user, action
                        ),
                    )
                else:
                    self.assertFalse(
                        self.policy.user_has_permission(user, action),
                        msg="User %s should not be able to %s, but can"
                        % (user, action),
                    )

    def assertUserInstancePermissionMatrix(self, instance, test_cases, actions=()):
        """
        Given a list of (user, can_change, can_delete, can_frobnicate) tuples
        (where 'frobnicate' is an unrecognised action not defined on the model),
        confirm that all tuples correctly represent permissions for that user on
        the given instance, as returned by user_has_permission_for_instance
        """
        if not actions:
            actions = ["change", "delete", "frobnicate"]
        for test_case in test_cases:
            user = test_case[0]
            expected_results = zip(actions, test_case[1:])

            for action, expected_result in expected_results:
                if expected_result:
                    self.assertTrue(
                        self.policy.user_has_permission_for_instance(
                            user, action, instance
                        ),
                        msg="User %s should be able to %s instance %s, but can't"
                        % (user, action, instance),
                    )
                else:
                    self.assertFalse(
                        self.policy.user_has_permission_for_instance(
                            user, action, instance
                        ),
                        msg="User %s should not be able to %s instance %s, but can"
                        % (user, action, instance),
                    )


class PermissionPolicyTestCase(PermissionPolicyTestUtils, WagtailTestUtils, TestCase):
    def setUp(self):
        # Permissions
        image_content_type = ContentType.objects.get_for_model(Image)
        add_image_permission = Permission.objects.get(
            content_type=image_content_type, codename="add_image"
        )
        change_image_permission = Permission.objects.get(
            content_type=image_content_type, codename="change_image"
        )
        delete_image_permission = Permission.objects.get(
            content_type=image_content_type, codename="delete_image"
        )

        # Groups
        image_adders_group = Group.objects.create(name="Image adders")
        image_adders_group.permissions.add(add_image_permission)

        image_changers_group = Group.objects.create(name="Image changers")
        image_changers_group.permissions.add(change_image_permission)

        # Users
        self.superuser = self.create_superuser(
            "superuser", "superuser@example.com", "password"
        )
        self.inactive_superuser = self.create_superuser(
            "inactivesuperuser", "inactivesuperuser@example.com", "password"
        )
        self.inactive_superuser.is_active = False
        self.inactive_superuser.save()

        # a user with add_image permission through the 'Image adders' group
        self.image_adder = self.create_user(
            "imageadder", "imageadder@example.com", "password"
        )
        self.image_adder.groups.add(image_adders_group)

        # a user with add_image permission through user_permissions
        self.oneoff_image_adder = self.create_user(
            "oneoffimageadder", "oneoffimageadder@example.com", "password"
        )
        self.oneoff_image_adder.user_permissions.add(add_image_permission)

        # a user that has add_image permission, but is inactive
        self.inactive_image_adder = self.create_user(
            "inactiveimageadder", "inactiveimageadder@example.com", "password"
        )
        self.inactive_image_adder.groups.add(image_adders_group)
        self.inactive_image_adder.is_active = False
        self.inactive_image_adder.save()

        # a user with change_image permission through the 'Image changers' group
        self.image_changer = self.create_user(
            "imagechanger", "imagechanger@example.com", "password"
        )
        self.image_changer.groups.add(image_changers_group)

        # a user with change_image permission through user_permissions
        self.oneoff_image_changer = self.create_user(
            "oneoffimagechanger", "oneoffimagechanger@example.com", "password"
        )
        self.oneoff_image_changer.user_permissions.add(change_image_permission)

        # a user that has change_image permission, but is inactive
        self.inactive_image_changer = self.create_user(
            "inactiveimagechanger", "inactiveimagechanger@example.com", "password"
        )
        self.inactive_image_changer.groups.add(image_changers_group)
        self.inactive_image_changer.is_active = False
        self.inactive_image_changer.save()

        # a user with delete_image permission through user_permissions
        self.oneoff_image_deleter = self.create_user(
            "oneoffimagedeleter", "oneoffimagedeleter@example.com", "password"
        )
        self.oneoff_image_deleter.user_permissions.add(delete_image_permission)

        # a user with no permissions
        self.useless_user = self.create_user(
            "uselessuser", "uselessuser@example.com", "password"
        )

        self.anonymous_user = AnonymousUser()

        # Images

        # an image owned by 'imageadder'
        self.adder_image = Image.objects.create(
            title="imageadder's image",
            file=get_test_image_file(),
            uploaded_by_user=self.image_adder,
        )

        # an image owned by 'uselessuser'
        self.useless_image = Image.objects.create(
            title="uselessuser's image",
            file=get_test_image_file(),
            uploaded_by_user=self.useless_user,
        )

        # an image with no owner
        self.anonymous_image = Image.objects.create(
            title="anonymous image",
            file=get_test_image_file(),
        )


class TestBlanketPermissionPolicy(PermissionPolicyTestCase):
    def setUp(self):
        super().setUp()
        self.policy = BlanketPermissionPolicy(Image)

        self.active_users = [
            self.superuser,
            self.image_adder,
            self.oneoff_image_adder,
            self.image_changer,
            self.oneoff_image_changer,
            self.oneoff_image_deleter,
            self.useless_user,
        ]
        self.all_users = self.active_users + [
            self.inactive_superuser,
            self.inactive_image_adder,
            self.inactive_image_changer,
            self.anonymous_user,
        ]

    def test_user_has_permission(self):
        # All users have permission to do everything
        self.assertUserPermissionMatrix(
            [(user, True, True, True, True) for user in self.all_users]
        )

    def test_user_has_any_permission(self):
        for user in self.all_users:
            self.assertTrue(
                self.policy.user_has_any_permission(user, ["add", "change"])
            )

    def test_users_with_permission(self):
        # all active users have permission
        users_with_add_permission = self.policy.users_with_permission("add")

        self.assertResultSetEqual(users_with_add_permission, self.active_users)

    def test_users_with_any_permission(self):
        # all active users have permission
        users_with_add_or_change_permission = self.policy.users_with_any_permission(
            ["add", "change"]
        )

        self.assertResultSetEqual(
            users_with_add_or_change_permission, self.active_users
        )

    def test_user_has_permission_for_instance(self):
        # All users have permission to do everything on any given instance
        self.assertUserInstancePermissionMatrix(
            self.adder_image, [(user, True, True, True) for user in self.all_users]
        )

    def test_user_has_any_permission_for_instance(self):
        for user in self.all_users:
            self.assertTrue(
                self.policy.user_has_any_permission_for_instance(
                    user, ["change", "delete"], self.adder_image
                )
            )

    def test_instances_user_has_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]

        # all users can edit all instances
        for user in self.all_users:
            self.assertResultSetEqual(
                self.policy.instances_user_has_permission_for(user, "change"),
                all_images,
            )

    def test_instances_user_has_any_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]

        for user in self.all_users:
            self.assertResultSetEqual(
                self.policy.instances_user_has_any_permission_for(
                    user, ["change", "delete"]
                ),
                all_images,
            )

    def test_users_with_permission_for_instance(self):
        # all active users have permission
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.useless_image
        )

        self.assertResultSetEqual(users_with_change_permission, self.active_users)

    def test_users_with_any_permission_for_instance(self):
        # all active users have permission
        users_with_change_or_del_permission = (
            self.policy.users_with_any_permission_for_instance(
                ["change", "delete"], self.useless_image
            )
        )

        self.assertResultSetEqual(
            users_with_change_or_del_permission, self.active_users
        )


class TestAuthenticationOnlyPermissionPolicy(PermissionPolicyTestCase):
    def setUp(self):
        super().setUp()
        self.policy = AuthenticationOnlyPermissionPolicy(Image)

    def test_user_has_permission(self):
        # All active authenticated users have permission to do everything;
        # inactive and anonymous users have permission to do nothing
        self.assertUserPermissionMatrix(
            [
                (self.superuser, True, True, True, True),
                (self.inactive_superuser, False, False, False, False),
                (self.image_adder, True, True, True, True),
                (self.oneoff_image_adder, True, True, True, True),
                (self.inactive_image_adder, False, False, False, False),
                (self.image_changer, True, True, True, True),
                (self.oneoff_image_changer, True, True, True, True),
                (self.inactive_image_changer, False, False, False, False),
                (self.oneoff_image_deleter, True, True, True, True),
                (self.useless_user, True, True, True, True),
                (self.anonymous_user, False, False, False, False),
            ]
        )

    def test_user_has_any_permission(self):
        self.assertTrue(
            self.policy.user_has_any_permission(self.superuser, ["add", "change"])
        )

        self.assertFalse(
            self.policy.user_has_any_permission(
                self.inactive_superuser, ["add", "change"]
            )
        )

        self.assertTrue(
            self.policy.user_has_any_permission(self.useless_user, ["add", "change"])
        )

        self.assertFalse(
            self.policy.user_has_any_permission(self.anonymous_user, ["add", "change"])
        )

    def test_users_with_permission(self):
        # all active users have permission
        users_with_add_permission = self.policy.users_with_permission("add")

        self.assertResultSetEqual(
            users_with_add_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
                self.useless_user,
            ],
        )

    def test_users_with_any_permission(self):
        # all active users have permission
        users_with_add_or_change_permission = self.policy.users_with_any_permission(
            ["add", "change"]
        )

        self.assertResultSetEqual(
            users_with_add_or_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
                self.useless_user,
            ],
        )

    def test_user_has_permission_for_instance(self):
        # Permissions for this policy are applied at the model level,
        # so rules for a specific instance will match rules for the
        # model as a whole
        self.assertUserInstancePermissionMatrix(
            self.adder_image,
            [
                (self.superuser, True, True, True),
                (self.inactive_superuser, False, False, False),
                (self.image_adder, True, True, True),
                (self.oneoff_image_adder, True, True, True),
                (self.inactive_image_adder, False, False, False),
                (self.image_changer, True, True, True),
                (self.oneoff_image_changer, True, True, True),
                (self.inactive_image_changer, False, False, False),
                (self.oneoff_image_deleter, True, True, True),
                (self.useless_user, True, True, True),
                (self.anonymous_user, False, False, False),
            ],
        )

    def test_user_has_any_permission_for_instance(self):
        # superuser has permission
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.superuser, ["change", "delete"], self.adder_image
            )
        )

        # inactive user has no permission
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.inactive_superuser, ["change", "delete"], self.adder_image
            )
        )

        # ordinary user has permission
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.useless_user, ["change", "delete"], self.adder_image
            )
        )

        # anonymous user has no permission
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.anonymous_user, ["change", "delete"], self.adder_image
            )
        )

    def test_instances_user_has_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.superuser, "change"),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.inactive_superuser, "change"
            ),
            no_images,
        )

        # the set of images editable by ordinary user includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.useless_user, "change"),
            all_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.anonymous_user, "change"
            ),
            no_images,
        )

    def test_instances_user_has_any_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.superuser, ["change", "delete"]
            ),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.inactive_superuser, ["change", "delete"]
            ),
            no_images,
        )

        # the set of images editable by ordinary user includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.useless_user, ["change", "delete"]
            ),
            all_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.anonymous_user, ["change", "delete"]
            ),
            no_images,
        )

    def test_users_with_permission_for_instance(self):
        # all active users have permission
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.useless_image
        )

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
                self.useless_user,
            ],
        )

    def test_users_with_any_permission_for_instance(self):
        # all active users have permission
        users_with_change_or_del_permission = (
            self.policy.users_with_any_permission_for_instance(
                ["change", "delete"], self.useless_image
            )
        )

        self.assertResultSetEqual(
            users_with_change_or_del_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
                self.useless_user,
            ],
        )


class TestModelPermissionPolicy(PermissionPolicyTestCase):
    def setUp(self):
        super().setUp()
        self.policy = ModelPermissionPolicy(Image)

    def test_user_has_permission(self):
        self.assertUserPermissionMatrix(
            [
                # Superuser has permission to do everything
                (self.superuser, True, True, True, True),
                # Inactive superuser can do nothing
                (self.inactive_superuser, False, False, False, False),
                # User with 'add' permission via group can only add
                (self.image_adder, True, False, False, False),
                # User with 'add' permission via user can only add
                (self.oneoff_image_adder, True, False, False, False),
                # Inactive user with 'add' permission can do nothing
                (self.inactive_image_adder, False, False, False, False),
                # User with 'change' permission via group can only change
                (self.image_changer, False, True, False, False),
                # User with 'change' permission via user can only change
                (self.oneoff_image_changer, False, True, False, False),
                # Inactive user with 'add' permission can do nothing
                (self.inactive_image_changer, False, False, False, False),
                # User with 'delete' permission can only delete
                (self.oneoff_image_deleter, False, False, True, False),
                # User with no permissions can do nothing
                (self.useless_user, False, False, False, False),
                # Anonymous user can do nothing
                (self.anonymous_user, False, False, False, False),
            ]
        )

    def test_user_has_any_permission(self):
        # Superuser can do everything
        self.assertTrue(
            self.policy.user_has_any_permission(self.superuser, ["add", "change"])
        )

        # Inactive superuser can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(
                self.inactive_superuser, ["add", "change"]
            )
        )

        # Only one of the permissions in the list needs to pass
        # in order for user_has_any_permission to return true
        self.assertTrue(
            self.policy.user_has_any_permission(self.image_adder, ["add", "change"])
        )
        self.assertTrue(
            self.policy.user_has_any_permission(
                self.oneoff_image_adder, ["add", "change"]
            )
        )
        self.assertTrue(
            self.policy.user_has_any_permission(self.image_changer, ["add", "change"])
        )

        # User with some permission, but not the ones in the list,
        # should return false
        self.assertFalse(
            self.policy.user_has_any_permission(self.image_changer, ["add", "delete"])
        )

        # Inactive user with the appropriate permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(
                self.inactive_image_adder, ["add", "delete"]
            )
        )

        # User with no permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(self.useless_user, ["add", "change"])
        )

        # Anonymous user can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(self.anonymous_user, ["add", "change"])
        )

    def test_users_with_permission(self):
        users_with_add_permission = self.policy.users_with_permission("add")

        self.assertResultSetEqual(
            users_with_add_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
            ],
        )

        users_with_change_permission = self.policy.users_with_permission("change")

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

    def test_users_with_any_permission(self):
        users_with_add_or_change_permission = self.policy.users_with_any_permission(
            ["add", "change"]
        )

        self.assertResultSetEqual(
            users_with_add_or_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        users_with_change_or_delete_permission = self.policy.users_with_any_permission(
            ["change", "delete"]
        )

        self.assertResultSetEqual(
            users_with_change_or_delete_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
            ],
        )

    def test_user_has_permission_for_instance(self):
        # Permissions for this policy are applied at the model level,
        # so rules for a specific instance will match rules for the
        # model as a whole
        self.assertUserInstancePermissionMatrix(
            self.adder_image,
            [
                (self.superuser, True, True, True),
                (self.inactive_superuser, False, False, False),
                (self.image_adder, False, False, False),
                (self.oneoff_image_adder, False, False, False),
                (self.inactive_image_adder, False, False, False),
                (self.image_changer, True, False, False),
                (self.oneoff_image_changer, True, False, False),
                (self.inactive_image_changer, False, False, False),
                (self.oneoff_image_deleter, False, True, False),
                (self.useless_user, False, False, False),
                (self.anonymous_user, False, False, False),
            ],
        )

    def test_user_has_any_permission_for_instance(self):
        # Superuser can do everything
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.superuser, ["change", "delete"], self.adder_image
            )
        )

        # Inactive superuser can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.inactive_superuser, ["change", "delete"], self.adder_image
            )
        )

        # Only one of the permissions in the list needs to pass
        # in order for user_has_any_permission to return true
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.image_changer, ["change", "delete"], self.adder_image
            )
        )
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.oneoff_image_changer, ["change", "delete"], self.adder_image
            )
        )

        # User with some permission, but not the ones in the list,
        # should return false
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.image_adder, ["change", "delete"], self.adder_image
            )
        )

        # Inactive user with the appropriate permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.inactive_image_changer, ["change", "delete"], self.adder_image
            )
        )

        # User with no permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.useless_user, ["change", "delete"], self.adder_image
            )
        )

        # Anonymous user can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.anonymous_user, ["change", "delete"], self.adder_image
            )
        )

    def test_instances_user_has_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.superuser, "change"),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.inactive_superuser, "change"
            ),
            no_images,
        )

        # given the relevant model permission at the group level, a user can edit all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.image_changer, "change"),
            all_images,
        )

        # given the relevant model permission at the user level, a user can edit all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.oneoff_image_changer, "change"
            ),
            all_images,
        )

        # a user with no permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.useless_user, "change"),
            no_images,
        )

        # an inactive user with the relevant permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.inactive_image_changer, "change"
            ),
            no_images,
        )

        # a user with permission, but not the matching one, can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.image_changer, "delete"),
            no_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.anonymous_user, "change"
            ),
            no_images,
        )

    def test_instances_user_has_any_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.superuser, ["change", "delete"]
            ),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.inactive_superuser, ["change", "delete"]
            ),
            no_images,
        )

        # given the relevant model permission at the group level, a user can edit all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.image_changer, ["change", "delete"]
            ),
            all_images,
        )

        # given the relevant model permission at the user level, a user can edit all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.oneoff_image_changer, ["change", "delete"]
            ),
            all_images,
        )

        # a user with no permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.useless_user, ["change", "delete"]
            ),
            no_images,
        )

        # an inactive user with the relevant permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.inactive_image_changer, ["change", "delete"]
            ),
            no_images,
        )

        # a user with permission, but not the matching one, can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.image_adder, ["change", "delete"]
            ),
            no_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.anonymous_user, ["change", "delete"]
            ),
            no_images,
        )

    def test_users_with_permission_for_instance(self):
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.useless_image
        )

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        users_with_delete_permission = self.policy.users_with_permission_for_instance(
            "delete", self.useless_image
        )

        self.assertResultSetEqual(
            users_with_delete_permission,
            [
                self.superuser,
                self.oneoff_image_deleter,
            ],
        )

    def test_users_with_any_permission_for_instance(self):
        users_with_change_or_del_permission = (
            self.policy.users_with_any_permission_for_instance(
                ["change", "delete"], self.useless_image
            )
        )

        self.assertResultSetEqual(
            users_with_change_or_del_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
                self.oneoff_image_deleter,
            ],
        )


class TestOwnershipPermissionPolicy(PermissionPolicyTestCase):
    def setUp(self):
        super().setUp()
        self.policy = OwnershipPermissionPolicy(
            Image, owner_field_name="uploaded_by_user"
        )

    def test_user_has_permission(self):
        self.assertUserPermissionMatrix(
            [
                # Superuser has permission to do everything
                (self.superuser, True, True, True, True),
                # Inactive superuser can do nothing
                (self.inactive_superuser, False, False, False, False),
                # User with 'add' permission via group can add,
                # and by extension, change and delete their own instances
                (self.image_adder, True, True, True, False),
                # User with 'add' permission via user can add,
                # and by extension, change and delete their own instances
                (self.oneoff_image_adder, True, True, True, False),
                # Inactive user with 'add' permission can do nothing
                (self.inactive_image_adder, False, False, False, False),
                # User with 'change' permission via group can change and delete but not add
                (self.image_changer, False, True, True, False),
                # User with 'change' permission via user can change and delete but not add
                (self.oneoff_image_changer, False, True, True, False),
                # Inactive user with 'change' permission can do nothing
                (self.inactive_image_changer, False, False, False, False),
                # 'delete' permission is ignored for this policy
                (self.oneoff_image_deleter, False, False, False, False),
                # User with no permission can do nothing
                (self.useless_user, False, False, False, False),
                # Anonymous user can do nothing
                (self.anonymous_user, False, False, False, False),
            ]
        )

    def test_user_has_any_permission(self):
        # Superuser can do everything
        self.assertTrue(
            self.policy.user_has_any_permission(self.superuser, ["add", "change"])
        )

        # Inactive superuser can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(
                self.inactive_superuser, ["add", "change"]
            )
        )

        # Only one of the permissions in the list needs to pass
        # in order for user_has_any_permission to return true
        self.assertTrue(
            self.policy.user_has_any_permission(self.image_changer, ["add", "change"])
        )
        self.assertTrue(
            self.policy.user_has_any_permission(
                self.oneoff_image_changer, ["add", "change"]
            )
        )

        # User with some permission, but not the ones in the list,
        # should return false
        self.assertFalse(
            self.policy.user_has_any_permission(
                self.oneoff_image_deleter, ["add", "change"]
            )
        )

        # Inactive user with the appropriate permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(
                self.inactive_image_changer, ["add", "delete"]
            )
        )

        # User with no permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(self.useless_user, ["add", "change"])
        )

        # Anonymous user can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission(self.anonymous_user, ["add", "change"])
        )

    def test_users_with_permission(self):
        users_with_add_permission = self.policy.users_with_permission("add")

        self.assertResultSetEqual(
            users_with_add_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
            ],
        )

        # users with add permission have change permission too (i.e. for their own images)
        users_with_change_permission = self.policy.users_with_permission("change")

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        # conditions for deletion are the same as for change; 'delete' permission
        # records in django.contrib.auth are ignored
        users_with_delete_permission = self.policy.users_with_permission("delete")

        self.assertResultSetEqual(
            users_with_delete_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        # non-standard permissions are only available to superusers
        users_with_frobnicate_permission = self.policy.users_with_permission(
            "frobnicate"
        )

        self.assertResultSetEqual(
            users_with_frobnicate_permission,
            [
                self.superuser,
            ],
        )

    def test_users_with_any_permission(self):
        users_with_add_or_change_permission = self.policy.users_with_any_permission(
            ["add", "change"]
        )

        self.assertResultSetEqual(
            users_with_add_or_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        users_with_add_or_frobnicate_permission = self.policy.users_with_any_permission(
            ["add", "frobnicate"]
        )

        self.assertResultSetEqual(
            users_with_add_or_frobnicate_permission,
            [
                self.superuser,
                self.image_adder,
                self.oneoff_image_adder,
            ],
        )

    def test_user_has_permission_for_instance(self):
        # Test permissions for an image owned by image_adder
        self.assertUserInstancePermissionMatrix(
            self.adder_image,
            [
                # superuser can do everything
                (self.superuser, True, True, True),
                # inactive superuser can do nothing
                (self.inactive_superuser, False, False, False),
                # image_adder can change and delete their own images,
                # but not perform custom actions
                (self.image_adder, True, True, False),
                # user with add permission cannot edit images owned by others
                (self.oneoff_image_adder, False, False, False),
                # inactive user with 'add' permission can do nothing
                (self.inactive_image_adder, False, False, False),
                # user with change permission can change and delete all images
                (self.image_changer, True, True, False),
                # likewise for change permission specified at the user level
                (self.oneoff_image_changer, True, True, False),
                # inactive user with 'change' permission can do nothing
                (self.inactive_image_changer, False, False, False),
                # delete permissions are ignored
                (self.oneoff_image_deleter, False, False, False),
                # user with no permissions can do nothing
                (self.useless_user, False, False, False),
                # anonymous user can do nothing
                (self.anonymous_user, False, False, False),
            ],
        )

        # Test permissions for an image owned by useless_user
        self.assertUserInstancePermissionMatrix(
            self.useless_image,
            [
                # superuser can do everything
                (self.superuser, True, True, True),
                # image_adder cannot edit images owned by others
                (self.image_adder, False, False, False),
                (self.oneoff_image_adder, False, False, False),
                # user with change permission can change and delete all images
                (self.image_changer, True, True, False),
                (self.oneoff_image_changer, True, True, False),
                # inactive users can do nothing
                (self.inactive_superuser, False, False, False),
                (self.inactive_image_adder, False, False, False),
                (self.inactive_image_changer, False, False, False),
                # delete permissions are ignored
                (self.oneoff_image_deleter, False, False, False),
                # user with no permissions can do nothing, even on images
                # they own
                (self.useless_user, False, False, False),
                # anonymous user can do nothing
                (self.anonymous_user, False, False, False),
            ],
        )

        # Instances with a null owner should always follow the same rules
        # as 'an instance owned by someone else'
        self.assertUserInstancePermissionMatrix(
            self.anonymous_image,
            [
                (self.superuser, True, True, True),
                (self.image_adder, False, False, False),
                (self.oneoff_image_adder, False, False, False),
                (self.image_changer, True, True, False),
                (self.oneoff_image_changer, True, True, False),
                (self.inactive_superuser, False, False, False),
                (self.inactive_image_adder, False, False, False),
                (self.inactive_image_changer, False, False, False),
                (self.oneoff_image_deleter, False, False, False),
                (self.useless_user, False, False, False),
                (self.anonymous_user, False, False, False),
            ],
        )

    def test_user_has_any_permission_for_instance(self):
        # Superuser can do everything
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.superuser, ["change", "delete"], self.adder_image
            )
        )

        # Inactive superuser can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.inactive_superuser, ["change", "delete"], self.adder_image
            )
        )

        # Only one of the permissions in the list needs to pass
        # in order for user_has_any_permission to return true
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.image_changer, ["change", "frobnicate"], self.adder_image
            )
        )
        self.assertTrue(
            self.policy.user_has_any_permission_for_instance(
                self.oneoff_image_changer, ["change", "frobnicate"], self.adder_image
            )
        )

        # User with some permission, but not the ones in the list,
        # should return false
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.oneoff_image_deleter, ["change", "delete"], self.adder_image
            )
        )

        # Inactive user with the appropriate permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.inactive_image_changer, ["change", "delete"], self.adder_image
            )
        )

        # User with no permissions can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.useless_user, ["change", "delete"], self.adder_image
            )
        )

        # Anonymous user can do nothing
        self.assertFalse(
            self.policy.user_has_any_permission_for_instance(
                self.anonymous_user, ["change", "delete"], self.adder_image
            )
        )

    def test_instances_user_has_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.superuser, "change"),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.inactive_superuser, "change"
            ),
            no_images,
        )

        # a user with 'add' permission can change their own images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.image_adder, "change"),
            [self.adder_image],
        )
        # a user with 'add' permission can also delete their own images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.image_adder, "delete"),
            [self.adder_image],
        )

        # a user with 'change' permission can change all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.image_changer, "change"),
            all_images,
        )

        # ditto for 'change' permission assigned at the user level
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.oneoff_image_changer, "change"
            ),
            all_images,
        )

        # an inactive user with the relevant permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.inactive_image_changer, "change"
            ),
            no_images,
        )

        # a user with no permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(self.useless_user, "change"),
            no_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_permission_for(
                self.anonymous_user, "change"
            ),
            no_images,
        )

    def test_instances_user_has_any_permission_for(self):
        all_images = [self.adder_image, self.useless_image, self.anonymous_image]
        no_images = []

        # the set of images editable by superuser includes all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.superuser, ["change", "delete"]
            ),
            all_images,
        )

        # the set of images editable by inactive superuser includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.inactive_superuser, ["change", "delete"]
            ),
            no_images,
        )

        # a user with 'add' permission can change/delete their own images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.image_adder, ["delete", "frobnicate"]
            ),
            [self.adder_image],
        )

        # a user with 'edit' permission can change/delete all images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.oneoff_image_changer, ["delete", "frobnicate"]
            ),
            all_images,
        )

        # a user with no permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.useless_user, ["change", "delete"]
            ),
            no_images,
        )

        # an inactive user with the relevant permission can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.inactive_image_changer, ["change", "delete"]
            ),
            no_images,
        )

        # a user with permission, but not the matching one, can edit no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.oneoff_image_deleter, ["change", "delete"]
            ),
            no_images,
        )

        # the set of images editable by anonymous user includes no images
        self.assertResultSetEqual(
            self.policy.instances_user_has_any_permission_for(
                self.anonymous_user, ["change", "delete"]
            ),
            no_images,
        )

    def test_users_with_permission_for_instance(self):
        # adder_image can be edited by its owner (who has add permission) and
        # all users with 'change' permission
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.adder_image
        )

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        # the same set of users can also delete
        users_with_delete_permission = self.policy.users_with_permission_for_instance(
            "delete", self.adder_image
        )

        self.assertResultSetEqual(
            users_with_delete_permission,
            [
                self.superuser,
                self.image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        # custom actions are available to superusers only
        users_with_delete_permission = self.policy.users_with_permission_for_instance(
            "frobnicate", self.adder_image
        )

        self.assertResultSetEqual(
            users_with_delete_permission,
            [
                self.superuser,
            ],
        )

        # useless_user can NOT edit their own image, because they do not have
        # 'add' permission
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.useless_image
        )

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

        # an image with no owner is treated as if it's owned by 'somebody else' -
        # i.e. users with 'change' permission can edit it
        users_with_change_permission = self.policy.users_with_permission_for_instance(
            "change", self.anonymous_image
        )

        self.assertResultSetEqual(
            users_with_change_permission,
            [
                self.superuser,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )

    def test_users_with_any_permission_for_instance(self):
        users_with_change_or_frob_permission = (
            self.policy.users_with_any_permission_for_instance(
                ["change", "frobnicate"], self.adder_image
            )
        )

        self.assertResultSetEqual(
            users_with_change_or_frob_permission,
            [
                self.superuser,
                self.image_adder,
                self.image_changer,
                self.oneoff_image_changer,
            ],
        )



from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.utils.functional import cached_property
from django.utils.translation import gettext as _


class WagtailPaginator(Paginator):
    num_page_buttons = 6

    @cached_property
    def model(self):
        return getattr(self.object_list, "model", None)

    @cached_property
    def verbose_name(self):
        if self.model:
            return self.model._meta.verbose_name
        return _("item")

    @cached_property
    def verbose_name_plural(self):
        if self.model:
            return self.model._meta.verbose_name_plural
        return _("items")

    @cached_property
    def items_count_label(self):
        if self.count == 1:
            return f"1 {self.verbose_name}"
        else:
            return f"{self.count} {self.verbose_name_plural}"

    def get_elided_page_range(self, page_number):
        """
        Provides a range of page numbers where the number of positions
        occupied by page numbers and ellipses is fixed to num_page_buttons.

        For example, if there are 10 pages where num_page_buttons is 6, the output will be:
        At page 1:  1 2 3 4 … 10
        At page 6:  1 … 6 7 … 10
        At page 10: 1 … 7 8 9 10

        The paginator will show the current page in the middle (odd number of buttons)
        or to the left side of the middle (even number of buttons).
        """

        try:
            number = self.validate_number(page_number)
        except PageNotAnInteger:
            number = 1
        except EmptyPage:
            number = self.num_pages

        if self.num_page_buttons < 5:
            # We provide no page range if fewer than 5 num_page_buttons.
            # This displays only "Previous" and "Next" buttons.
            return []

        # Provide all page numbers if fewer than num_page_buttons.
        if self.num_pages <= self.num_page_buttons:
            yield from self.page_range
            return

        # These thresholds are the maximum number of buttons
        # that can be shown on the start or end of the page range
        # before the middle part of the range expands.
        # For even num_page_buttons values both thresholds are the same.
        # For odd num_page_buttons values the start threshold is one more than the end threshold.
        end_threshold = self.num_page_buttons // 2
        start_threshold = end_threshold + (self.num_page_buttons % 2)

        # Show the first page.
        yield 1

        # Show middle pages.
        if number <= start_threshold:
            # Result: 1 [ 2 3 4 … ] 10
            yield from range(2, self.num_page_buttons - 1)
            yield self.ELLIPSIS
        elif number < self.num_pages - end_threshold:
            # Result: 1 [ … 5 6* 7 … ] 10
            # 4 spaces are occupied by first/last page numbers and ellipses
            middle_size = self.num_page_buttons - 4
            offset = (middle_size - 1) // 2
            yield self.ELLIPSIS
            yield from range(number - offset, number + middle_size - offset)
            yield self.ELLIPSIS
        else:
            # Result: 1 [ … 7 8 9 ] 10
            yield self.ELLIPSIS
            yield from range(
                self.num_pages - (self.num_page_buttons - 3), self.num_pages
            )

        # Show the last page.
        yield self.num_pages
